@@ -59,20 +59,10 @@ from firebirdsql.fbcore import ( __version__, apilevel, threadsafety,
 
 import firebirdsql.services
 
-def connect(dsn=None, user=None, password=None, role=None, host=None, 
-            database=None, charset=DEFAULT_CHARSET, port=3050,
-            timeout=None, isolation_level=None, use_unicode=None):
-    return Connection(dsn=dsn, user=user, password=password,
-                role=role, host=host, 
-                database=database, charset=charset, port=port,
-                timeout=timeout, isolation_level=isolation_level,
-                use_unicode=use_unicode)
+def connect(**kwargs):
+    return Connection(**kwargs)
 
-def create_database(dsn=None, user=None, password=None, host=None, role=None,
-            database=None, charset=DEFAULT_CHARSET, port=3050, page_size=4096,
-            timeout=None, isolation_level=None, use_unicode=None):
-    return Connection(dsn=dsn, user=user, password=password,
-                role=role, host=host, 
-                database=database, charset=charset, port=port, 
-                page_size=page_size, timeout=timeout,
-                isolation_level=isolation_level, use_unicode=use_unicode)
+def create_database(**kwargs):
+    if not 'page_size' in kwargs:
+        kwargs['page_size'] = 4096
+    return Connection(**kwargs)
