@@ -6,16 +6,17 @@
 #
 # Python DB-API 2.0 module for Firebird.
 ##############################################################################
-import sys, os, socket
+import sys
 import xdrlib, time, datetime, decimal, struct
 import itertools
 from firebirdsql.fberrmsgs import messages
 from firebirdsql import (DatabaseError, InternalError, OperationalError,
     ProgrammingError, IntegrityError, DataError, NotSupportedError,)
 from firebirdsql.consts import *
-from firebirdsql.wireprotocol import (WireProtocol,
-    bytes_to_bint, bytes_to_int, bint_to_bytes, int_to_bytes, byte_to_int,
-    INFO_SQL_SELECT_DESCRIBE_VARS,)
+from firebirdsql.utils import *
+from firebirdsql.wireprotocol import WireProtocol, INFO_SQL_SELECT_DESCRIBE_VARS
+from firebirdsql.socketstream import SocketStream
+from firebirdsql.xsqlvar import XSQLVAR, calc_blr, parse_select_items, parse_xsqlda
 
 try:
     from collections import Mapping
@@ -25,21 +26,7 @@ except ImportError:
     from UserDict import DictMixin as Mapping
     HAS_MAPPING = False
 
-try:
-    import fcntl
-except ImportError:
-    def setcloexec(sock):
-        pass
-else:
-    def setcloexec(sock):
-        """Set FD_CLOEXEC property on a file descriptors
-        """
-        fd = sock.fileno()
-        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
-        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
-
-
-__version__ = '0.8.2'
+__version__ = '0.8.3'
 apilevel = '2.0'
 threadsafety = 1
 paramstyle = 'qmark'
@@ -89,265 +76,6 @@ NUMBER = DBAPITypeObject(int, decimal.Decimal)
 DATETIME = DBAPITypeObject(datetime.datetime, datetime.date, datetime.time)
 ROWID = DBAPITypeObject()
 
-#------------------------------------------------------------------------------
-class XSQLVAR:
-    type_length = {
-        SQL_TYPE_VARYING: -1,
-        SQL_TYPE_SHORT: 4,
-        SQL_TYPE_LONG: 4,
-        SQL_TYPE_FLOAT: 4,
-        SQL_TYPE_TIME: 4,
-        SQL_TYPE_DATE: 4,
-        SQL_TYPE_DOUBLE: 8,
-        SQL_TYPE_TIMESTAMP: 8,
-        SQL_TYPE_BLOB: 8,
-        SQL_TYPE_ARRAY: 8,
-        SQL_TYPE_QUAD: 8,
-        SQL_TYPE_INT64: 8,
-        SQL_TYPE_BOOLEAN: 1,
-        }
-
-    type_display_length = {
-        SQL_TYPE_VARYING: -1,
-        SQL_TYPE_SHORT: 6,
-        SQL_TYPE_LONG: 11,
-        SQL_TYPE_FLOAT: 17,
-        SQL_TYPE_TIME: 11,
-        SQL_TYPE_DATE: 10,
-        SQL_TYPE_DOUBLE: 17,
-        SQL_TYPE_TIMESTAMP: 22,
-        SQL_TYPE_BLOB: 0,
-        SQL_TYPE_ARRAY: -1,
-        SQL_TYPE_QUAD: 20,
-        SQL_TYPE_INT64: 20,
-        SQL_TYPE_BOOLEAN: 5,
-        }
-
-    def __init__(self, bytes_to_str):
-        self.bytes_to_str = bytes_to_str
-        self.sqltype = None
-        self.sqlscale = None
-        self.sqlsubtype = None
-        self.sqllen = None
-        self.null_ok = None
-        self.fieldname = ''
-        self.relname = ''
-        self.ownname = ''
-        self.aliasname = ''
-
-    def io_length(self):
-        sqltype = self.sqltype
-        if sqltype == SQL_TYPE_TEXT:
-            return self.sqllen
-        else:
-            return self.type_length[sqltype]
-
-    def display_length(self):
-        sqltype = self.sqltype
-        if sqltype == SQL_TYPE_TEXT:
-            return self.sqllen
-        else:
-            return self.type_display_length[sqltype]
-
-    def precision(self):
-        return None
-
-    def __str__(self):
-        s  = '[' + str(self.sqltype) + ',' + str(self.sqlscale) + ',' \
-                + str(self.sqlsubtype) + ',' + str(self.sqllen)  + ',' \
-                + str(self.null_ok) + ',' + self.fieldname + ',' \
-                + self.relname + ',' + self.ownname + ',' \
-                + self.aliasname + ']'
-        return s
-
-    def _parse_date(self, raw_value):
-        "Convert raw data to datetime.date"
-        nday = bytes_to_bint(raw_value) + 678882
-        century = (4 * nday -1) // 146097
-        nday = 4 * nday - 1 - 146097 * century
-        day = nday // 4
-
-        nday = (4 * day + 3) // 1461
-        day = 4 * day + 3 - 1461 * nday
-        day = (day + 4) // 4
-
-        month = (5 * day -3) // 153
-        day = 5 * day - 3 - 153 * month
-        day = (day + 5) // 5
-        year = 100 * century + nday
-        if month < 10:
-            month += 3
-        else:
-            month -= 9
-            year += 1
-        return year, month, day
-
-    def _parse_time(self, raw_value):
-        "Convert raw data to datetime.time"
-        n = bytes_to_bint(raw_value)
-        s = n // 10000
-        m = s // 60
-        h = m // 60
-        m = m % 60
-        s = s % 60
-        return (h, m, s, (n % 10000) * 100)
-
-    def value(self, raw_value):
-        if self.sqltype == SQL_TYPE_TEXT:
-            return self.bytes_to_str(raw_value).rstrip()
-        elif self.sqltype == SQL_TYPE_VARYING:
-            return self.bytes_to_str(raw_value)
-        elif self.sqltype in (SQL_TYPE_SHORT, SQL_TYPE_LONG, SQL_TYPE_INT64):
-            n = bytes_to_bint(raw_value)
-            if self.sqlscale:
-                return decimal.Decimal(str(n) + 'e' + str(self.sqlscale))
-            else:
-                return n
-        elif self.sqltype == SQL_TYPE_DATE:
-            yyyy, mm, dd = self._parse_date(raw_value)
-            return datetime.date(yyyy, mm, dd)
-        elif self.sqltype == SQL_TYPE_TIME:
-            h, m, s, ms = self._parse_time(raw_value)
-            return datetime.time(h, m, s, ms)
-        elif self.sqltype == SQL_TYPE_TIMESTAMP:
-            yyyy, mm, dd = self._parse_date(raw_value[:4])
-            h, m, s, ms = self._parse_time(raw_value[4:])
-            return datetime.datetime(yyyy, mm, dd, h, m, s, ms)
-        elif self.sqltype == SQL_TYPE_FLOAT:
-            return struct.unpack('!f', raw_value)[0]
-        elif self.sqltype == SQL_TYPE_DOUBLE:
-            return struct.unpack('!d', raw_value)[0]
-        elif self.sqltype == SQL_TYPE_BOOLEAN:
-            return True if byte_to_int(raw_value[0]) != 0 else False
-        else:
-            return raw_value
-
-
-sqltype2blr = {
-    SQL_TYPE_DOUBLE: [27],
-    SQL_TYPE_FLOAT: [10],
-    SQL_TYPE_D_FLOAT: [11],
-    SQL_TYPE_DATE: [12],
-    SQL_TYPE_TIME: [13],
-    SQL_TYPE_TIMESTAMP: [35],
-    SQL_TYPE_BLOB: [9, 0],
-    SQL_TYPE_ARRAY: [9, 0],
-    SQL_TYPE_BOOLEAN: [23],
-    }
-
-def calc_blr(xsqlda):
-    "Calculate  BLR from XSQLVAR array."
-    ln = len(xsqlda) *2
-    blr = [5, 2, 4, 0, ln & 255, ln >> 8]
-    for x in xsqlda:
-        sqltype = x.sqltype
-        if sqltype == SQL_TYPE_VARYING:
-            blr += [37, x.sqllen & 255, x.sqllen >> 8]
-        elif sqltype == SQL_TYPE_TEXT:
-            blr += [14, x.sqllen & 255, x.sqllen >> 8]
-        elif sqltype == SQL_TYPE_LONG:
-            blr += [8, x.sqlscale]
-        elif sqltype == SQL_TYPE_SHORT:
-            blr += [7, x.sqlscale]
-        elif sqltype == SQL_TYPE_INT64:
-            blr += [16, x.sqlscale]
-        elif sqltype == SQL_TYPE_QUAD:
-            blr += [9, x.sqlscale]
-        else:
-            blr += sqltype2blr[sqltype]
-        blr += [7, 0]   # [blr_short, 0]
-    blr += [255, 76]    # [blr_end, blr_eoc]
-
-    # x.sqlscale value shoud be negative, so b convert to range(0, 256)
-    return bytes(256 + b if b < 0 else b for b in blr)
-
-def parse_select_items(buf, xsqlda, connection):
-    index = 0
-    i = 0
-    item = byte_to_int(buf[i])
-    while item != isc_info_end:
-        if item == isc_info_sql_sqlda_seq:
-            l = bytes_to_int(buf[i+1:i+3])
-            index = bytes_to_int(buf[i+3:i+3+l])
-            xsqlda[index-1] = XSQLVAR(connection.bytes_to_ustr
-                                        if connection.use_unicode
-                                        else connection.bytes_to_str)
-            xsqlda[index-1] = XSQLVAR(connection.bytes_to_str)
-            i = i + 3 + l
-        elif item == isc_info_sql_type:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].sqltype = bytes_to_int(buf[i+3:i+3+l]) & ~ 1
-            i = i + 3 + l
-        elif item == isc_info_sql_sub_type:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].sqlsubtype = bytes_to_int(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_sql_scale:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].sqlscale = bytes_to_int(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_sql_length:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].sqllen = bytes_to_int(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_sql_null_ind:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].null_ok = bytes_to_int(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_sql_field:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].fieldname = connection.bytes_to_str(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_sql_relation:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].relname = connection.bytes_to_str(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_sql_owner:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].ownname = connection.bytes_to_str(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_sql_alias:
-            l = bytes_to_int(buf[i+1:i+3])
-            xsqlda[index-1].aliasname = connection.bytes_to_str(buf[i+3:i+3+l])
-            i = i + 3 + l
-        elif item == isc_info_truncated:
-            return index    # return next index
-        elif item == isc_info_sql_describe_end:
-            i = i + 1
-        else:
-            print('\t', item, 'Invalid item [%02x] ! i=%d' % (buf[i], i))
-            i = i + 1
-        item = byte_to_int(buf[i])
-    return -1   # no more info
-
-def parse_xsqlda(buf, connection, stmt_handle):
-    xsqlda = []
-    stmt_type = None
-    i = 0
-    while i < len(buf):
-        if buf[i:i+3] == bytes([isc_info_sql_stmt_type,0x04,0x00]):
-            stmt_type = bytes_to_int(buf[i+3:i+7])
-            i += 7
-        elif buf[i:i+2] == bytes([isc_info_sql_select, isc_info_sql_describe_vars]):
-            i += 2
-            l = bytes_to_int(buf[i:i+2])
-            i += 2
-            col_len = bytes_to_int(buf[i:i+l])
-            xsqlda = [None] * col_len
-            next_index = parse_select_items(buf[i+l:], xsqlda, connection)
-            while next_index > 0:   # more describe vars
-                connection._op_info_sql(stmt_handle,
-                            bytes([isc_info_sql_sqlda_start, 2])
-                                + int_to_bytes(next_index, 2)
-                                + INFO_SQL_SELECT_DESCRIBE_VARS)
-                (h, oid, buf) = connection._op_response()
-                assert buf[:2] == bytes([0x04,0x07])
-                l = bytes_to_int(buf[2:4])
-                assert bytes_to_int(buf[4:4+l]) == col_len
-                next_index = parse_select_items(buf[4+l:], xsqlda, connection)
-        else:
-            break
-    return stmt_type, xsqlda
 
 class PreparedStatement:
     def __init__(self, cur, sql, explain_plan=False):
@@ -657,8 +385,7 @@ class EventConduit(WireProtocol):
             self.event_names[name] = 0
         self.timeout = timeout
         (h, port, family, ip_address) = self.connection._op_connect_request()
-        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        self.sock.connect((ip_address, port))
+        self.sock = SocketStream(ip_address, port, timeout)
         self.connection.last_event_id += 1
         self.event_id = self.connection.last_event_id
 
@@ -691,18 +418,9 @@ class EventConduit(WireProtocol):
         self.connection._op_cancel_events(self.event_id)
         (h, oid, buf) = self.connection._op_response()
         self.sock.close()
+        self.sock = None
 
 class Connection(WireProtocol):
-    def uid(self):
-        if sys.platform == 'win32':
-            user = os.environ['USERNAME']
-            hostname = os.environ['COMPUTERNAME']
-        else:
-            user = os.environ.get('USER', '')
-            hostname = socket.gethostname()
-        return bytes([1] + [len(user)] + [ord(c) for c in user]
-                + [4] + [len(hostname)] + [ord(c) for c in hostname] + [6, 0])
-
     def cursor(self):
         if self._transaction is None:
             self.begin()
@@ -727,6 +445,9 @@ class Connection(WireProtocol):
             self._transaction.rollback(retaining=retaining, savepoint=savepoint)
 
     def execute_immediate(self, query):
+        if self._transaction is None:
+            self._transaction = Transaction(self)
+            self._transaction.begin()
         self._op_exec_immediate(
             self._transaction.trans_handle, query=query)
         (h, oid, buf) = self._op_response()
@@ -734,7 +455,8 @@ class Connection(WireProtocol):
     def __init__(self, dsn=None, user=None, password=None, role=None, host=None,
                     database=None, charset=DEFAULT_CHARSET, port=3050,
                     page_size=None, is_services=False, cloexec=False,
-                    timeout=None, isolation_level=None, use_unicode=None):
+                    timeout=None, isolation_level=None, use_unicode=None,
+                    connect_version=2, use_srp=False, wire_crypt=False):
         if dsn:
             i = dsn.find(':')
             if i < 0:
@@ -767,17 +489,13 @@ class Connection(WireProtocol):
         else:
             self.isolation_level = int(isolation_level)
         self.use_unicode = use_unicode
+        self.connect_version = connect_version
         self.last_event_id = 0
 
         self._transaction = None
+        self.sock = SocketStream(self.hostname, self.port, self.timeout, cloexec)
 
-        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        if cloexec:
-            setcloexec(self.sock)
-        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
-        self.sock.connect((self.hostname, self.port))
-
-        self._op_connect()
+        self._op_connect(use_srp, wire_crypt)
         try:
             self._op_accept()
         except OperationalError:
@@ -967,6 +685,9 @@ class Connection(WireProtocol):
         if self.sock:
             self.close()
 
+    def is_disconnect(self):
+        return self.sock is None
+
 class Transaction:
     def __init__(self, connection, tpb=None):
         self._connection = connection
