@@ -6,14 +6,23 @@
 #
 # Python DB-API 2.0 module for Firebird. 
 ##############################################################################
+import sys
 import os
+import socket
 import xdrlib, time, datetime, decimal, struct, select
+try:
+    import crypt
+except ImportError: # Not posix
+    crypt = None
 from firebirdsql.fberrmsgs import messages
 from firebirdsql import (DisconnectByPeer,
     DatabaseError, InternalError, OperationalError,
     ProgrammingError, IntegrityError, DataError, NotSupportedError,
 )
 from firebirdsql.consts import *
+from firebirdsql.utils import *
+from firebirdsql import srp
+from firebirdsql.arc4 import Arc4
 
 DEBUG = False
 
@@ -58,69 +67,8 @@ def wire_operation(fn):
         return r
     return f
 
-def bytes_to_bint(b, u=False):           # Read as big endian
-    if u:
-        fmtmap = {1: 'B', 2: '>H', 4: '>L', 8: '>Q'}
-    else:
-        fmtmap = {1: 'b', 2: '>h', 4: '>l', 8: '>q'}
-    fmt = fmtmap.get(len(b))
-    if fmt is None:
-        raise InternalError
-    return struct.unpack(fmt, b)[0]
 
-def bytes_to_int(b, u=False):            # Read as little endian.
-    if u:
-        fmtmap = {1: 'B', 2: '<H', 4: '<L', 8: '<Q'}
-    else:
-        fmtmap = {1: 'b', 2: '<h', 4: '<l', 8: '<q'}
-    fmt = fmtmap.get(len(b))
-    if fmt is None:
-        raise InternalError
-    return struct.unpack(fmt, b)[0]
-
-def bint_to_bytes(val, nbytes): # Convert int value to big endian bytes.
-    v = abs(val)
-    b = []
-    for n in range(nbytes):
-        b.append((v >> (8*(nbytes - n - 1)) & 0xff))
-    if val < 0:
-        for i in range(nbytes):
-            b[i] = ~b[i] + 256
-        b[-1] += 1
-        for i in range(nbytes):
-            if b[nbytes -i -1] == 256:
-                b[nbytes -i -1] = 0
-                b[nbytes -i -2] += 1
-    return bytes(b)
-
-def int_to_bytes(val, nbytes):  # Convert int value to little endian bytes.
-    v = abs(val)
-    b = []
-    for n in range(nbytes):
-        b.append((v >> (8 * n)) & 0xff)
-    if val < 0:
-        for i in range(nbytes):
-            b[i] = ~b[i] + 256
-        b[0] += 1
-        for i in range(nbytes):
-            if b[i] == 256:
-                b[i] = 0
-                b[i+1] += 1
-    return bytes(b)
-
-def byte_to_int(b):
-    "byte to int"
-    if PYTHON_MAJOR_VER == 3:
-        return b
-    else:
-        return ord(b)
-
-def send_channel(sock, b):
-    n = 0
-    while (n < len(b)):
-        n += sock.send(b[n:])
-
-class WireProtocol:
+class WireProtocol(object):
     buffer_length = 1024
 
     op_connect = 1
@@ -167,7 +115,19 @@ class WireProtocol:
     op_service_info = 84
     op_service_start = 85
     op_rollback_retaining = 86
-
+    # FB3
+    op_update_account_info = 87
+    op_authenticate_user = 88
+    op_partial = 89
+    op_trusted_auth = 90
+    op_cancel = 91
+    op_cont_auth = 92
+    op_ping = 93
+    op_accept_data = 94
+    op_abort_aux_connection = 95
+    op_crypt = 96
+    op_crypt_key_callback = 97
+    op_cond_accept = 98
 
     def recv_channel(self, nbytes, word_alignment=False):
         n = nbytes
@@ -177,7 +137,7 @@ class WireProtocol:
         while n:
             if (os.name != 'java'
                 and self.timeout is not None
-                and select.select([self.sock], [], [], self.timeout)[0] == []):
+                and select.select([self.sock._sock], [], [], self.timeout)[0] == []):
                 break
             b = self.sock.recv(n)
             if not b:
@@ -343,23 +303,83 @@ class WireProtocol:
         blr += bytes([255, 76])    # [blr_end, blr_eoc]
         return blr, values
 
+    def uid(self, use_srp, wire_crypt):
+        def pack_cnct_param(k, v):
+            if k != CNCT_specific_data:
+                return bytes([k] + [len(v)]) + v
+            # specific_data split per 254 bytes
+            b = b''
+            i = 0
+            while len(v) > 254:
+                b += bytes([k, 255, i]) + v[:254]
+                v = v[254:]
+                i += 1
+            b += bytes([k, len(v)+1, i]) + v
+            return b
+
+        if sys.platform == 'win32':
+            user = os.environ['USERNAME']
+            hostname = os.environ['COMPUTERNAME']
+        else:
+            user = os.environ.get('USER', '')
+            hostname = socket.gethostname()
+        r = b''
+        if self.connect_version == 3:
+            if use_srp:
+                self.client_public_key, self.client_private_key = \
+                    srp.client_seed(self.str_to_bytes(self.user.upper()),
+                                    self.str_to_bytes(self.password))
+                self.plugin_name = b'Srp'
+                self.plugin_list = b'Srp, Legacy_Auth'
+                specific_data = bytes_to_hex(srp.long2bytes(self.client_public_key))
+            else:
+                self.plugin_name = b'Legacy_Auth'
+                self.plugin_list = b'Legacy_Auth'
+                specific_data = self.str_to_bytes(
+                                        crypt.crypt(self.password, '9z')[2:])
+    
+            if wire_crypt:
+                client_crypt = int_to_bytes(1, 4)
+            else:
+                client_crypt = int_to_bytes(0, 4)
+    
+            r += pack_cnct_param(CNCT_login, self.str_to_bytes(self.user.upper()))
+            r += pack_cnct_param(CNCT_plugin_name, self.plugin_name)
+            r += pack_cnct_param(CNCT_plugin_list, self.plugin_list)
+            r += pack_cnct_param(CNCT_specific_data, specific_data)
+            r += pack_cnct_param(CNCT_client_crypt, client_crypt)
+        r += pack_cnct_param(CNCT_user, self.str_to_bytes(user))
+        r += pack_cnct_param(CNCT_host, self.str_to_bytes(hostname))
+        r += pack_cnct_param(CNCT_user_verification, b'')
+        return r
 
     @wire_operation
-    def _op_connect(self):
+    def _op_connect(self, use_srp=False, wire_crypt=False):
+        arch_type = 36
+        min_arch_type = 0
+        max_arch_type = 5
+        protocol_version_understood_count = 4
+        more_protocol = hex_to_bytes('ffff800b00000001000000000000000500000004ffff800c00000001000000000000000500000006ffff800d00000001000000000000000500000008')
+        if not use_srp and crypt is None:
+            self.connect_version = 2
+        if self.connect_version ==2:
+            max_arch_type = 3
+            protocol_version_understood_count = 1
+            more_protocol = b''
         p = xdrlib.Packer()
         p.pack_int(self.op_connect)
         p.pack_int(self.op_attach)
-        p.pack_int(2)   # CONNECT_VERSION2
-        p.pack_int(1)   # Arch type (Generic = 1)
+        p.pack_int(self.connect_version)
+        p.pack_int(arch_type)
         p.pack_string(self.str_to_bytes(self.filename if self.filename else ''))
-        p.pack_int(1)   # Protocol version understood count.
-        p.pack_bytes(self.uid())
+        p.pack_int(protocol_version_understood_count)
+        p.pack_bytes(self.uid(use_srp, wire_crypt))
         p.pack_int(10)  # PROTOCOL_VERSION10
-        p.pack_int(1)   # Arch type (Generic = 1)
-        p.pack_int(2)   # Min type
-        p.pack_int(3)   # Max type
+        p.pack_int(1)   # Protocol Arch type (Generic = 1)
+        p.pack_int(min_arch_type)
+        p.pack_int(max_arch_type)
         p.pack_int(2)   # Preference weight
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer()+more_protocol)
 
     @wire_operation
     def _op_create(self, page_size=4096):
@@ -369,21 +389,22 @@ class WireProtocol:
         dpb += bytes([isc_dpb_lc_ctype, len(s)]) + s
         s = self.str_to_bytes(self.user)
         dpb += bytes([isc_dpb_user_name, len(s)]) + s
-        s = self.str_to_bytes(self.password)
-        dpb += bytes([isc_dpb_password, len(s)]) + s
+        if self.connect_version == 2:
+            s = self.str_to_bytes(self.password)
+            dpb += bytes([isc_dpb_password, len(s)]) + s
         if self.role:
             s = self.str_to_bytes(self.role)
             dpb += bytes([isc_dpb_sql_role_name, len(s)]) + s
         dpb += bytes([isc_dpb_sql_dialect, 4]) + int_to_bytes(3, 4)
-        dpb += bytes([isc_dpb_force_write, 4]) + bint_to_bytes(1, 4)
-        dpb += bytes([isc_dpb_overwrite, 4]) + bint_to_bytes(1, 4)
+        dpb += bytes([isc_dpb_force_write, 4]) + int_to_bytes(1, 4)
+        dpb += bytes([isc_dpb_overwrite, 4]) + int_to_bytes(1, 4)
         dpb += bytes([isc_dpb_page_size, 4]) + int_to_bytes(page_size, 4)
         p = xdrlib.Packer()
         p.pack_int(self.op_create)
         p.pack_int(0)                       # Database Object ID
         p.pack_string(self.str_to_bytes(self.filename))
         p.pack_bytes(dpb)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_accept(self):
@@ -392,14 +413,82 @@ class WireProtocol:
             b = self.recv_channel(4)
         if bytes_to_bint(b) == self.op_reject:
             raise OperationalError('Connection is rejected', None, None)
-        assert bytes_to_bint(b) == self.op_accept
+
+        op_code = bytes_to_bint(b)
         b = self.recv_channel(12)
         up = xdrlib.Unpacker(b)
-        assert up.unpack_int() == 10
-        assert  up.unpack_int() == 1
-        assert up.unpack_int() == 3
+        protocol_version = up.unpack_int()
+        protocol_arch = up.unpack_int()
+        minimum_type =  up.unpack_int()
         up.done()
 
+        if op_code == self.op_cond_accept or op_code == self.op_accept_data:
+            read_length = 0
+
+            ln = bytes_to_bint(self.recv_channel(4))
+            data = self.recv_channel(ln)
+            read_length += 4 + ln
+            if read_length % 4:
+                self.recv_channel(4 - read_length % 4) # padding
+                read_length += 4 - read_length % 4
+
+            ln = bytes_to_bint(self.recv_channel(4))
+            self.plugin_name = self.recv_channel(ln)
+            read_length += 4 + ln
+            if read_length % 4:
+                self.recv_channel(4 - read_length % 4) # padding
+                read_length += 4 - read_length % 4
+
+            is_authenticated = bytes_to_bint(self.recv_channel(4))
+            read_length += 4
+            ln = bytes_to_bint(self.recv_channel(4))
+            keys = self.recv_channel(ln)
+            read_length += 4 + ln
+            if read_length % 4:
+                self.recv_channel(4 - read_length % 4) # padding
+                read_length += 4 - read_length % 4
+
+            if self.plugin_name == b'Legacy_Auth' and is_authenticated == 0:
+                raise OperationalError('Unauthorized', 0, 0)
+
+            if self.plugin_name == b'Srp':
+                ln = bytes_to_int(data[:2])
+                self.server_salt = data[2:ln+2]
+                self.server_public_key = data[4+ln:]
+#                print('server_salt=', self.server_salt)
+#                print('server_public_key=', self.server_public_key)
+                self.server_public_key = srp.bytes2long(self.server_public_key)
+
+                self.client_proof, self.auth_key = srp.client_proof(
+                                            self.str_to_bytes(self.user),
+                                            self.str_to_bytes(self.password),
+                                            self.server_salt,
+                                            self.server_public_key,
+                                            self.client_public_key,
+                                            self.client_private_key)
+#                print('client_proof=', bytes_to_hex(self.client_proof))
+                # send op_cont_auth
+                p = xdrlib.Packer()
+                p.pack_int(self.op_cont_auth)
+                p.pack_string(bytes_to_hex(self.client_proof))
+                p.pack_bytes(self.plugin_name)
+                p.pack_bytes(self.plugin_list)
+                p.pack_bytes(b'')
+                self.sock.send(p.get_buffer())
+                (h, oid, buf) = self._op_response()
+
+                # op_crypt: plugin[Arc4] key[Symmetric]
+                p = xdrlib.Packer()
+                p.pack_int(self.op_crypt)
+                p.pack_string('Arc4')
+                p.pack_string('Symmetric')
+                self.sock.send(p.get_buffer() + b'\00\00\00')   # 3bytes padding
+                p.sock.set_translator(Arc4(self.auth_key), Arc4(self.auth_key))
+                (h, oid, buf) = self._op_response()
+        else:
+            assert op_code == self.op_accept
+            self.connect_version = 2
+
     @wire_operation
     def _op_attach(self):
         dpb = bytes([1])
@@ -407,8 +496,9 @@ class WireProtocol:
         dpb += bytes([isc_dpb_lc_ctype, len(s)]) + s
         s = self.str_to_bytes(self.user)
         dpb += bytes([isc_dpb_user_name, len(s)]) + s
-        s = self.str_to_bytes(self.password)
-        dpb += bytes([isc_dpb_password, len(s)]) + s
+        if self.connect_version == 2:
+            s = self.str_to_bytes(self.password)
+            dpb += bytes([isc_dpb_password, len(s)]) + s
         if self.role:
             s = self.str_to_bytes(self.role)
             dpb += bytes([isc_dpb_sql_role_name, len(s)]) + s
@@ -417,14 +507,14 @@ class WireProtocol:
         p.pack_int(0)                       # Database Object ID
         p.pack_string(self.str_to_bytes(self.filename))
         p.pack_bytes(dpb)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_drop_database(self):
         p = xdrlib.Packer()
         p.pack_int(self.op_drop_database)
         p.pack_int(self.db_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_service_attach(self):
@@ -439,7 +529,7 @@ class WireProtocol:
         p.pack_int(0)
         p.pack_string(self.str_to_bytes('service_mgr'))
         p.pack_bytes(dpb)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_service_info(self, param, item, buffer_length=512):
@@ -451,7 +541,7 @@ class WireProtocol:
         p.pack_bytes(param)
         p.pack_bytes(item)
         p.pack_int(buffer_length)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_service_start(self, param):
@@ -460,14 +550,14 @@ class WireProtocol:
         p.pack_int(self.db_handle)
         p.pack_int(0)
         p.pack_bytes(param)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_service_detach(self):
         p = xdrlib.Packer()
         p.pack_int(self.op_service_detach)
         p.pack_int(self.db_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_info_database(self, b):
@@ -477,7 +567,7 @@ class WireProtocol:
         p.pack_int(0)
         p.pack_bytes(b)
         p.pack_int(self.buffer_length)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_transaction(self, tpb):
@@ -485,42 +575,42 @@ class WireProtocol:
         p.pack_int(self.op_transaction)
         p.pack_int(self.db_handle)
         p.pack_bytes(tpb)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_commit(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_commit)
         p.pack_int(trans_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_commit_retaining(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_commit_retaining)
         p.pack_int(trans_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_rollback(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_rollback)
         p.pack_int(trans_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_rollback_retaining(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_rollback_retaining)
         p.pack_int(trans_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_allocate_statement(self):
         p = xdrlib.Packer()
         p.pack_int(self.op_allocate_statement)
         p.pack_int(self.db_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_info_transaction(self, trans_handle, b):
@@ -530,7 +620,7 @@ class WireProtocol:
         p.pack_int(0)
         p.pack_bytes(b)
         p.pack_int(self.buffer_length)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_free_statement(self, stmt_handle, mode):
@@ -538,7 +628,7 @@ class WireProtocol:
         p.pack_int(self.op_free_statement)
         p.pack_int(stmt_handle)
         p.pack_int(mode)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_prepare_statement(self, stmt_handle, trans_handle, query, option_items=bytes([])):
@@ -551,7 +641,7 @@ class WireProtocol:
         p.pack_string(self.str_to_bytes(query))
         p.pack_bytes(desc_items)
         p.pack_int(self.buffer_length)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_info_sql(self, stmt_handle, vars):
@@ -561,7 +651,7 @@ class WireProtocol:
         p.pack_int(0)
         p.pack_bytes(vars)
         p.pack_int(self.buffer_length)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_execute(self, stmt_handle, trans_handle, params):
@@ -574,13 +664,13 @@ class WireProtocol:
             p.pack_bytes(bytes([]))
             p.pack_int(0)
             p.pack_int(0)
-            send_channel(self.sock, p.get_buffer())
+            self.sock.send(p.get_buffer())
         else:
             (blr, values) = self.params_to_blr(trans_handle, params)
             p.pack_bytes(blr)
             p.pack_int(0)
             p.pack_int(1)
-            send_channel(self.sock, p.get_buffer() + values)
+            self.sock.send(p.get_buffer() + values)
 
     @wire_operation
     def _op_execute2(self, stmt_handle, trans_handle, params, output_blr):
@@ -593,18 +683,18 @@ class WireProtocol:
             p.pack_bytes(bytes([]))
             p.pack_int(0)
             p.pack_int(0)
-            send_channel(self.sock, p.get_buffer())
+            self.sock.send(p.get_buffer())
         else:
             (blr, values) = self.params_to_blr(trans_handle, params)
             p.pack_bytes(blr)
             p.pack_int(0)
             p.pack_int(1)
-            send_channel(self.sock, p.get_buffer() + values)
+            self.sock.send(p.get_buffer() + values)
 
         p = xdrlib.Packer()
         p.pack_bytes(output_blr)
         p.pack_int(0)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_exec_immediate(self, trans_handle, query):
@@ -617,7 +707,7 @@ class WireProtocol:
         p.pack_string(self.str_to_bytes(query))
         p.pack_bytes(desc_items)
         p.pack_int(self.buffer_length)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_fetch(self, stmt_handle, blr):
@@ -627,7 +717,7 @@ class WireProtocol:
         p.pack_bytes(blr)
         p.pack_int(0)
         p.pack_int(400)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_fetch_response(self, stmt_handle, xsqlda):
@@ -666,14 +756,14 @@ class WireProtocol:
         p = xdrlib.Packer()
         p.pack_int(self.op_detach)
         p.pack_int(self.db_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_open_blob(self, blob_id, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_open_blob)
         p.pack_int(trans_handle)
-        send_channel(self.sock, p.get_buffer() + blob_id)
+        self.sock.send(p.get_buffer() + blob_id)
 
     @wire_operation
     def _op_create_blob2(self, trans_handle):
@@ -683,7 +773,7 @@ class WireProtocol:
         p.pack_int(trans_handle)
         p.pack_int(0)
         p.pack_int(0)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_get_segment(self, blob_handle):
@@ -692,7 +782,7 @@ class WireProtocol:
         p.pack_int(blob_handle)
         p.pack_int(self.buffer_length)
         p.pack_int(0)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_put_segment(self, blob_handle, b):
@@ -701,7 +791,7 @@ class WireProtocol:
         p.pack_int(blob_handle)
         p.pack_int(len(b))
         p.pack_int(len(b))
-        send_channel(self.sock, p.get_buffer() + b)
+        self.sock.send(p.get_buffer() + b)
 
     @wire_operation
     def _op_batch_segments(self, blob_handle, seg_data):
@@ -712,7 +802,7 @@ class WireProtocol:
         p.pack_int(ln + 2)
         p.pack_int(ln + 2)
         pad_length = ((4-(ln+2)) & 3)
-        send_channel(self.sock, p.get_buffer() 
+        self.sock.send(p.get_buffer() 
                 + int_to_bytes(ln, 2) + seg_data + bytes([0])*pad_length)
 
     @wire_operation
@@ -720,7 +810,7 @@ class WireProtocol:
         p = xdrlib.Packer()
         p.pack_int(self.op_close_blob)
         p.pack_int(blob_handle)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_que_events(self, event_names, ast, args, event_id):
@@ -736,7 +826,7 @@ class WireProtocol:
         p.pack_int(ast)
         p.pack_int(args)
         p.pack_int(event_id)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_cancel_events(self, event_id):
@@ -744,7 +834,7 @@ class WireProtocol:
         p.pack_int(self.op_cancel_events)
         p.pack_int(self.db_handle)
         p.pack_int(event_id)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
     @wire_operation
     def _op_connect_request(self):
@@ -753,7 +843,7 @@ class WireProtocol:
         p.pack_int(1)    # async
         p.pack_int(self.db_handle)
         p.pack_int(0)
-        send_channel(self.sock, p.get_buffer())
+        self.sock.send(p.get_buffer())
 
         b = self.recv_channel(4)
         while bytes_to_bint(b) == self.op_dummy:
